@@ -30,12 +30,120 @@
 #include "functions.h"
 
 #include "get.h"
+#include "register.h"
 #include "mp.h"
 #include "mpmath.h"
 #include "display.h"
 #include "ce_parser.h"
 #include "ui.h"
 
+enum button_flags {
+    NUMBER       = (1 << 3),   /* Number button */
+    FUNC         = (1 << 6),   /* Function */
+    PREFIXOP     = (1 << 15),  /* Unary prefix operation */
+};
+
+typedef struct {
+    int id;
+    char *symname;           /* Expression function name */
+    enum button_flags flags; /* Misc flags */
+} Function;
+
+// FIXME: Sort this list
+/* Note that none of these strings can be translated as the parser expects them to be correct */
+/* id, symname flags */
+static Function functions[NFUNCTIONS] = {
+{ FN_0,                 "0", NUMBER },
+{ FN_1,                 "1", NUMBER },
+{ FN_2,                 "2", NUMBER },    
+{ FN_3,                 "3", NUMBER },
+{ FN_4,                 "4", NUMBER },
+{ FN_5,                 "5", NUMBER },
+{ FN_6,                 "6", NUMBER },
+{ FN_7,                 "7", NUMBER },
+{ FN_8,                 "8", NUMBER },
+{ FN_9,                 "9", NUMBER },
+{ FN_A,                 "A", NUMBER },
+{ FN_B,                 "B", NUMBER },    
+{ FN_C,                 "C", NUMBER },
+{ FN_D,                 "D", NUMBER },
+{ FN_E,                 "E", NUMBER },
+{ FN_F,                 "F", NUMBER },
+{ FN_NUMERIC_POINT,     ".", NUMBER },
+{ FN_CALCULATE,         NULL, 0 },
+{ FN_CLEAR,             NULL, 0 },
+{ FN_CLEAR_ENTRY,       NULL, 0 },
+{ FN_START_BLOCK,       "(", 0 },
+{ FN_END_BLOCK,         ")", 0 },
+{ FN_ADD,               "+", 0 },
+{ FN_SUBTRACT,          "-", 0 },
+{ FN_MULTIPLY,          "*", 0 },
+{ FN_DIVIDE,            "/", 0 },
+{ FN_BACKSPACE,         NULL, 0 },
+{ FN_DELETE,            NULL, 0 },
+{ FN_CHANGE_SIGN,       NULL, 0 },
+{ FN_INTEGER,           "Int", FUNC },
+{ FN_FRACTION,          "Frac", FUNC },
+{ FN_PERCENTAGE,        "%", 0 },
+{ FN_SQUARE,            "^2", 0 },
+{ FN_SQUARE_ROOT,       "Sqrt", FUNC },
+{ FN_RECIPROCAL,        NULL, 0 },
+{ FN_E_POW_X,           "e^", PREFIXOP },
+{ FN_10_POW_X,          "10^", PREFIXOP },       
+{ FN_2_POW_X,           "2^", PREFIXOP },
+{ FN_X_POW_Y,           "^", 0 },
+{ FN_X_POW_Y_INV,       "^(1/(", 0 },
+{ FN_FACTORIAL,         "!", 0 },
+{ FN_RANDOM,            "Rand", 0 },
+{ FN_SIN,               "Sin", FUNC },
+{ FN_SINH,              "Sinh", FUNC },
+{ FN_ASIN,              "Asin", FUNC },
+{ FN_ASINH,             "Asinh", FUNC },
+{ FN_COS,               "Cos", FUNC },
+{ FN_COSH,              "Cosh", FUNC },
+{ FN_ACOS,              "Acos", FUNC },
+{ FN_ACOSH,             "Acosh", FUNC },
+{ FN_TAN,               "Tan", FUNC },
+{ FN_TANH,              "Tanh", FUNC },
+{ FN_ATAN,              "Atan", FUNC },
+{ FN_TAN,               "Atanh", FUNC },
+{ FN_NATURAL_LOGARITHM, "Ln", FUNC },
+{ FN_LOGARITHM,         "Log", FUNC },
+{ FN_LOGARITHM2,        "Log2", FUNC },
+{ FN_ABSOLUTE_VALUE,    "Abs", FUNC },
+{ FN_MASK_16,           "u16", FUNC },            
+{ FN_MASK_32,           "u32", FUNC },
+{ FN_MODULUS_DIVIDE,    " Mod ", 0 },
+{ FN_EXPONENTIAL,       "e", 0 },
+{ FN_NOT,               "~", 0 },
+{ FN_OR,                " OR ", 0 },
+{ FN_AND,               " AND ", 0 },       
+{ FN_XOR,               " XOR ", 0 },
+{ FN_XNOR,              " XNOR ", 0 },
+{ FN_TOGGLE_BIT,        NULL, 0 },
+{ FN_FINC_CTRM,         "Ctrm", 0 },
+{ FN_FINC_DDB,          "Ddb", 0 },
+{ FN_FINC_FV,           "Fv", 0 },
+{ FN_FINC_GPM,          "Gpm", 0 },
+{ FN_FINC_PMT,          "Pmt", 0 },
+{ FN_FINC_PV,           "Pv", 0 },
+{ FN_FINC_RATE,         "Rate", 0 },
+{ FN_FINC_SLN,          "Sln", 0 },
+{ FN_FINC_SYD ,         "Syd", 0 },
+{ FN_FINC_TERM,         "Term", 0 },
+{ FN_SHIFT,             NULL, 0 },
+{ FN_STORE,             NULL, 0 },
+{ FN_RECALL,            NULL, 0 },
+{ FN_EXCHANGE,          NULL, 0 },
+{ FN_SET_ACCURACY,      NULL, 0 },
+{ FN_SET_BASE,          NULL, 0 },
+{ FN_SET_NUMBERTYPE,    NULL, 0 },
+{ FN_SET_TRIG_TYPE,     NULL, 0 },
+{ FN_UNDO,              NULL, 0 },
+{ FN_REDO,              NULL, 0 },
+{ FN_CONSTANT,          NULL, 0 },
+{ FN_FUNCTION,          NULL, 0 }
+};
 
 static void
 clear_undo_history(void)
@@ -61,16 +169,12 @@ do_accuracy(int value)     /* Set display accuracy. */
 static void
 do_function(int index)      /* Perform a user defined function. */
 {
-    char *str;
     int ret;
 
     assert(index >= 0);
     assert(index <= 9);
 
-    str = v->fun_vals[index];
-    assert(str);
-    ret = ce_udf_parse(str);
-
+    ret = ce_udf_parse(function_get_value(index));
     if (!ret) {
         ui_set_statusbar("", "");
     } else {
@@ -138,8 +242,8 @@ do_exchange(int index)
         ui_set_statusbar(_("No sane value to store"),
                          "gtk-dialog-error");
     } else {
-        mp_set_from_mp(v->MPmvals[index], MPtemp);
-        mp_set_from_mp(MPexpr, v->MPmvals[index]);
+        register_get(index, MPtemp);
+        register_set(index, MPexpr);
         mp_set_from_mp(MPtemp, display_get_answer(&v->display));
         display_set_string(&v->display, "Ans", -1);
         display_refresh(&v->display);
@@ -178,10 +282,14 @@ do_numtype(enum num_type n)   /* Set number display type. */
 static void
 do_sto(int index)
 {
-    if (display_is_usable_number(&v->display, v->MPmvals[index])) {
+    int temp[MP_SIZE];
+    
+    if (display_is_usable_number(&v->display, temp)) {
         ui_set_statusbar(_("No sane value to store"),
                          "gtk-dialog-error");
     }
+    else
+        register_set(index, temp);
 
     ui_make_registers();
 }
@@ -192,13 +300,15 @@ do_expression(int function, int arg, int cursor)
 {
     char buf[MAXLINE];
     int *ans;
+    int enabled;
+    guint64 bit_value;
     
-    switch (buttons[function].id) {
-        case KEY_UNDO:
+    switch (functions[function].id) {
+        case FN_UNDO:
             display_pop(&v->display);
             return;
 
-        case KEY_REDO:
+        case FN_REDO:
             display_unpop(&v->display);
             return;
 
@@ -215,74 +325,96 @@ do_expression(int function, int arg, int cursor)
 
     /* Starting a number after a calculation clears the display */
     if (display_is_result(&v->display)) {
-        if (buttons[function].flags & NUMBER) {
+        if (functions[function].flags & NUMBER) {
             display_clear(&v->display);
         }
     }
 
-    switch (buttons[function].id) {
-        case KEY_CLEAR:
-        case KEY_CLEAR_ENTRY:
+    switch (functions[function].id) {
+        case FN_CLEAR:
+        case FN_CLEAR_ENTRY:
             display_clear(&v->display);
             ui_set_error_state(FALSE);
             MPstr_to_num("0", 10, ans);
             break;
 
-        case KEY_SHIFT:
+        case FN_SHIFT:
             do_shift(arg);
             return;
 
-        case KEY_SET_ACCURACY:
+        case FN_SET_ACCURACY:
             do_accuracy(arg);
             return;
 
-        case KEY_SET_BASE:
+        case FN_SET_BASE:
             do_base(arg);
             return;
+        
+        case FN_SET_TRIG_TYPE:
+            v->ttype = arg;
+            set_resource(R_TRIG, Rtstr[arg]);
+            return;
 
-        case KEY_SET_NUMBERTYPE:
+        case FN_SET_NUMBERTYPE:
             do_numtype(arg);
             return;        
         
-        case KEY_FUNCTION:
+        case FN_FUNCTION:
             do_function(arg);
             return;
 
-        case KEY_STORE:
+        case FN_STORE:
             do_sto(arg);
             return;
 
-        case KEY_EXCHANGE:
+        case FN_EXCHANGE:
             do_exchange(arg);
             return;
 
-        case KEY_RECALL:
+        case FN_RECALL:
             SNPRINTF(buf, MAXLINE, "R%d", arg);
             display_insert(&v->display, buf);
             break;
 
-        case KEY_CONSTANT:
-            make_number(buf, MAXLINE, v->MPcon_vals[arg], v->base, FALSE);
+        case FN_CONSTANT:
+            make_number(buf, MAXLINE, constant_get_value(arg), v->base, FALSE);
             display_insert(&v->display, buf);
             break;
 
-        case KEY_BACKSPACE:
+        case FN_BACKSPACE:
             display_backspace(&v->display);
             break;
         
-        case KEY_DELETE:
+        case FN_DELETE:
             display_delete(&v->display);
             break;
 
-        case KEY_CHANGE_SIGN:
+        case FN_CHANGE_SIGN:
             display_surround(&v->display, "-(", ")");
             break;
 
-        case KEY_RECIPROCAL:
+        case FN_RECIPROCAL:
             display_surround(&v->display, "1/(", ")");
             break;
 
-        case KEY_CALCULATE:
+        case FN_TOGGLE_BIT:
+            if (display_get_unsigned_integer(&v->display, &bit_value)) {
+                char buf[MAX_DISPLAY];
+                int MP[MP_SIZE];
+
+                bit_value ^= (1LL << (63 - arg));
+    
+                /* FIXME: Convert to since we don't support setting MP numbers from 64 bit integers */
+                SNPRINTF(buf, MAX_DISPLAY, "%llu", bit_value);
+                MPstr_to_num(buf, 10, MP);
+
+                /* FIXME: Set as string as display_set_number doesn't store correctly */
+                make_number(buf, MAX_DISPLAY, MP, v->base, FALSE);
+                display_set_string(&v->display, buf, -1);
+            }
+            break;
+
+        case FN_CALCULATE:
             /* If showing a result display the calculation that caused
              * this result */
             /* TODO: Work out why two undo steps are required and why
@@ -340,24 +472,28 @@ do_expression(int function, int arg, int cursor)
             }
             break;
 
-        case KEY_NUMERIC_POINT:
+        case FN_NUMERIC_POINT:
             display_insert(&v->display, v->radix);
             break;
 
         default:
             /* If display is a number then perform functions on that number */
-            if (buttons[function].flags & (PREFIXOP | FUNC) && display_is_result(&v->display)) {
-                SNPRINTF(buf, MAXLINE, "%s(", buttons[function].symname);
+            if (functions[function].flags & (PREFIXOP | FUNC) && display_is_result(&v->display)) {
+                SNPRINTF(buf, MAXLINE, "%s(", functions[function].symname);
                 display_surround(&v->display, buf, ")");
             } else {
-                if (buttons[function].flags & FUNC) {
-                    SNPRINTF(buf, MAXLINE, "%s(", buttons[function].symname);
+                if (functions[function].flags & FUNC) {
+                    SNPRINTF(buf, MAXLINE, "%s(", functions[function].symname);
                     display_insert(&v->display, buf);
                 } else {
-                    display_insert(&v->display, buttons[function].symname);
+                    display_insert(&v->display, functions[function].symname);
                 }
             }
             break;
     }
+
     display_refresh(&v->display);
+    
+    enabled = display_get_unsigned_integer(&v->display, &bit_value);
+    ui_set_bitfield(enabled, bit_value);
 }
